/******************************************************************************
 *                             Core-extended                                  *
 *                                                                            *
 * Copyright (C) 2008- Jane Street Holding, LLC                               *
 *    Contact: opensource@janestreet.com                                      *
 *    WWW: http://www.janestreet.com/ocaml                                    *
 *                                                                            *
 *                                                                            *
 * This library is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public                 *
 * License as published by the Free Software Foundation; either               *
 * version 2 of the License, or (at your option) any later version.           *
 *                                                                            *
 * This library is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this library; if not, write to the Free Software        *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                            *
 ******************************************************************************/

/* Core_unix support functions written in C. */

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <caml/mlvalues.h>
#include <caml/unixsupport.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/threads.h>

/*
  If you want to turn on debugging you may use:

  #define fork_side_assert(v) assert(v)

  Note that assert uses non async-signal-safe functions. Do not leave this on
  in any production code
 */

#define fork_side_assert(ignore) ((void) 0)

#define SYSCALL(x)                              \
  while ((x) == -1) {                           \
    if (errno != EINTR) {                       \
      report_errno_on_pipe (err_fd,errno);      \
    }                                           \
  }                                             \

#define NONINTR(x)                              \
  while ((x) == -1){ assert(errno == EINTR); }  \

/* Copy an ocaml string array in a c string array terminated by
   a null pointer the result need to be free'd with a stat_free
   It is a copy of cstringvect in the ocaml unix's module.
 */
static char ** copy_stringvect(const value arg)
{
  char ** res;
  mlsize_t size, i;

  size = Wosize_val(arg);
  res = (char **) caml_stat_alloc((size + 1) * sizeof(char *));
  for (i = 0; i < size; i++) res[i] = String_val(Field(arg, i));
  res[size] = NULL;
  return res;
}

#ifdef __GNUC__
/* Giving Gcc as much info as possible */
static void report_errno_on_pipe (const int fd,const int my_err) __attribute__((noreturn));
#endif

/*
  Write an int to an fd.
  This function is designed to be used on the fork and therefor only uses
  async-signal-safe functions.
*/
static void report_errno_on_pipe (const int fd,const int my_err) {
  long unsigned int offset = 0;
  ssize_t out_chars;
  int ret;
  while (offset < sizeof(int)) {;
    switch (out_chars=write (fd,(char *) &my_err + offset,sizeof(int) - offset/sizeof(char))) {
    case -1:
      fork_side_assert (errno==EINTR);
      continue;
    default:
      offset += out_chars;
    }
  }
  fork_side_assert (offset == sizeof(int));
  ret = fsync(fd);
  fork_side_assert (ret);
  _exit(254);
}

/*
  Returns 0 if there was no errno printed on the pipe and -1 if there was one.
  Sets my_errno to the errno received.
*/
static int errno_from_pipe (const int fd,const int *my_errno) {
   ssize_t in_chars;
   long unsigned int offset = 0;
   while (1) {
     switch (in_chars=read(fd,
                           ((char *) my_errno) + offset,
                           sizeof(int) - (offset/sizeof(char))))
       {
       case -1 :
         assert (errno==EINTR);
         continue;
         /* The fd was closed with nothing written to it; no error */
       case 0:
         if (offset == 0) {
           /* The fd was closed with nothing written to it; no error */
           return 0;
         };
         assert (offset == sizeof(int));
         return -1;
       default:
         offset += in_chars;
       }
   };
}

static void set_cloexec (const int fd) {
  int flags,res;
  flags = fcntl(fd, F_GETFD);
  assert (flags != -1);
  flags |= FD_CLOEXEC;
  res=fcntl(fd, F_SETFD, flags);
  assert (res != -1);
}



CAMLprim value extended_ml_create_process
/* Returns the pid of the created process. If there's an error on the forked side the
   process exits with 254 and prints the error to v_stderr */
(value v_prog, /* Program name */
 value v_args, /* Full list of args passed to executable... */
 value v_stdin, /* Fd to connect to the forked stdin... */
 value v_stdout,
 value v_stderr,
 value v_fds_to_close, /* FDs to close on the forked side*/
 value v_working_dir, /* An fd to the directory we want to chdir too. -1 for none. */
 value v_env /* An ocaml (string array option) containing the environment to set*/
 )
{
  CAMLparam5(v_prog, v_args, v_stdin, v_stdout, v_stderr);
  CAMLxparam3(v_fds_to_close,v_working_dir,v_env);
  int stdin_fd = Int_val (v_stdin);
  int stdout_fd = Int_val (v_stdout);
  int stderr_fd = Int_val (v_stderr);
  char** envp  = NULL;
  sigset_t empty;
  int pfd[2]; /* The pipe used to report errors.. */
  int my_errno,status;
  int forked_error;


  int i;

  /* It's ok to hold pointers into the O'Caml heap, since the memory
     space gets duplicated upon the fork, during which we keep the
     O'Caml lock. */
  char* prog = String_val(v_prog);
  int working_dir = Int_val(v_working_dir);

  pid_t child_pid;
  int err_fd;

  char** args;


  /* We use a pipe to report errors on the forked side */
  if (pipe(pfd) == -1) uerror("extended_ml_create_process::pipe",Nothing);

  /* Set both side of the pipe close_on_exec... */
  set_cloexec(pfd[0]);
  set_cloexec(pfd[1]);

  args = copy_stringvect(v_args);

  if (Is_block(v_env))
    envp = copy_stringvect(Field(v_env,0));

  /* This function deliberately doesn't release the O'Caml lock (i.e. it
     doesn't call caml_enter_blocking_section) during the fork.  This is
     because we hold pointers into the ML heap across a fork, and
     releasing the lock immediately before the fork could theoretically
     cause the GC to run and move blocks before the fork duplicates the
     memory space. */
  switch (child_pid = fork()) {
  case -1:
    my_errno = errno;
    caml_stat_free(args);
    if (envp)
      caml_stat_free(envp);
    NONINTR(close(pfd[0]));
    NONINTR(close(pfd[1]));
    unix_error(my_errno,"extended_ml_create_process: fork failed", Nothing);
  case 0:
    err_fd = pfd[1];

    /* Child process.
       Since we've just lost all of our threads we need to be very careful
       not to call any function that might use a thread lock. This includes
       malloc,setenv and stdio functions... This is stated in the POSIX norm as:

       If a multi-threaded process calls fork(), the new process shall contain a
       replica of the calling thread and its entire address space, possibly
       including the states of mutexes and other resources. Consequently, to
       avoid errors, the child process may only execute async-signal-safe
       operations until such time as one of the exec functions is called.

       [http://pubs.opengroup.org/onlinepubs/009695399/functions/fork.html]

       We also need to use _exit instead of [exit] because we do not want
       [at_exit] registered functions to be called.
     */

    /* Reset the sigmask to get rid of the inherited one */
    sigemptyset (&empty);
    sigprocmask (SIG_SETMASK, &empty, (sigset_t *) 0);

    /* Close all the descriptors we want to get rid of. */
    for (i = Wosize_val(v_fds_to_close) - 1;i >= 0;i--) {
      SYSCALL(close(Int_val(Field(v_fds_to_close, i))));
    }

    /* Just in case any of the pipes' file descriptors are 0, 1 or 2
       (not inconceivable, especially when running as a daemon),
       duplicate all three descriptors we need in the child to fresh
       descriptors before duplicating them onto stdin, stdout and stderr.

       This will ensure that there is one and only one copy of the file
       descriptors passed as arguments with id's higher than 2.

       F_DUPFD cannot get EINTR so we'll go only once through the
       loop
    */
    SYSCALL(stdin_fd = fcntl(stdin_fd,F_DUPFD,3));
    SYSCALL(stdout_fd= fcntl(stdout_fd,F_DUPFD,3));
    SYSCALL(stderr_fd= fcntl(stderr_fd,F_DUPFD,3));

    /* We must dup2 the descriptors back in place... */
    SYSCALL(dup2(stdin_fd,0));
    SYSCALL(dup2(stdout_fd,1));
    SYSCALL(dup2(stderr_fd,2));

    /* And close the old fds... */
    SYSCALL(close(stdin_fd));
    SYSCALL(close(stdout_fd));
    SYSCALL(close(stderr_fd));

    if (working_dir != -1) {
      SYSCALL(fchdir(working_dir))
    }

    if (envp) {
      /* path lookups should be done on the parent side of the fork so no
         execvp*/
      SYSCALL(execve(prog, args,envp));
    }else {
      SYSCALL(execv(prog, args));
    }

    /* Raise an error?*/

  default : /* Parent process */

    caml_enter_blocking_section(); /* C side cleanup and looking for errors */
    NONINTR(close (pfd[1])); /* Close unused write end */

    forked_error = (errno_from_pipe(pfd[0],&my_errno));
    NONINTR(close (pfd[0]));

    if (forked_error)
      NONINTR(waitpid(child_pid, &status, WUNTRACED));

    caml_leave_blocking_section();

    /* Caml side cleanup */
    caml_stat_free(args);
    if (envp)
      caml_stat_free(envp);

    /* Returning the result */
    if  (forked_error)
      unix_error(my_errno,"extended_ml_create_process::forked_side" , Nothing);

    /* Reading the pipe.. */
    CAMLreturn(Val_int(child_pid));
  }
}

CAMLprim value extended_ml_create_process_bc(value *argv, int argn)
{
  if (argn != 8) {
    caml_failwith("Unix.ml_create_process_bc got the wrong number of \
     arguments. This is due to an error in the FFI.");
  }
  return
    extended_ml_create_process(
      argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6],
      argv[7]);
}
