(******************************************************************************
 *                             Core-extended                                  *
 *                                                                            *
 * Copyright (C) 2008- Jane Street Holding, LLC                               *
 *    Contact: opensource@janestreet.com                                      *
 *    WWW: http://www.janestreet.com/ocaml                                    *
 *                                                                            *
 *                                                                            *
 * This library is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public                 *
 * License as published by the Free Software Foundation; either               *
 * version 2 of the License, or (at your option) any later version.           *
 *                                                                            *
 * This library is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this library; if not, write to the Free Software        *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                            *
 ******************************************************************************)

open Core.Std

let default_lockfile_dir = "/var/tmp"

let command ?(lockfile_dir=default_lockfile_dir) ~name main =
  let lock_file = lockfile_dir ^/ name ^ ".lock" in
  let name_hum = sprintf "the %s service" name in
  let start_main () =
    let release_parent = Daemon.daemonize_wait () in
    (* lock file created after [daemonize_wait] so that *child* pid is written
       to the lock file rather than the parent pid *)
    if Lock_file.create lock_file (* this writes our pid in the file *) then begin
      (* we release the daemon's parent *after* the lock file is created
         so that any error messages during lock file creation happen
         prior to severing the daemon's connection to std{out,err} *)
      release_parent ();
      (* unix automatically handles unlocking the lock file, but it is
         our responsibility to delete it *)
      at_exit (fun () -> Unix.unlink lock_file);
      main ()
    end;
    0
  in
  let check_lock_file () =
    if Lock_file.is_locked lock_file then begin
      let pid = Int.t_of_sexp (Sexp.load_sexp lock_file) in
      `Running_with_pid pid
    end else
      `Not_running
  in
  let status_main () =
    begin match check_lock_file () with
    | `Running_with_pid pid -> printf "%s is running with pid %d\n%!" name_hum pid
    | `Not_running -> printf "%s is not running\n%!" name_hum
    end;
    0
  in
  let stop_aux ~stop_signal =
    let was_not_running () = 
      eprintf "%s was not running\n%!" name_hum;
      `Was_not_running
    in
    match check_lock_file () with
    | `Not_running -> was_not_running ()
    | `Running_with_pid pid ->
      let timeout_span = Time.Span.of_sec 10. in
      let deadline = Time.add (Time.now ()) timeout_span in
      match Signal.send stop_signal ~pid with
      | `No_such_process -> was_not_running ()
      | `Ok ->
        let still_alive () =
          match Signal.send Signal.zero ~pid with
          | `Ok -> true
          | `No_such_process -> false
        in
        let rec wait_loop () =
          if Time.(>=) (Time.now ()) deadline then begin
            eprintf "failed to observe %s die after %s\n%!" name_hum
              (Time.Span.to_string timeout_span);
            `Did_not_die
          end else if still_alive () then begin
            Time.pause (Time.Span.of_sec 0.2);
            wait_loop ()
          end else
            `Died
        in
        wait_loop ()
  in
  let stop_main ~stop_signal =
    match stop_aux ~stop_signal with
    | `Was_not_running | `Did_not_die -> 1
    | `Died -> 0
  in
  let restart_main ~stop_signal =
    match stop_aux ~stop_signal with
    | `Did_not_die -> 1
    | `Was_not_running | `Died -> start_main ()
  in
  let summary ~verb = sprintf "%s %s" verb name_hum in
  let assert_no_anons anons =
    match anons with
    | [] -> ()
    | anons ->
      failwithf "expected 0 anonymous arguments but found %d\n%!"
        (List.length anons) ()
  in
  let base_cmd ~verb main =
    Command.create (fun () -> exit (main ()))
      ~summary:(summary ~verb)
      ~usage_arg:""
      ~init:Fn.id
      ~flags:[]
      ~final:(fun () anons -> assert_no_anons anons)
  in
  let stop_cmd ~verb main =
    Command.create (fun stop_signal -> exit (main ~stop_signal))
      ~summary:(summary ~verb)
      ~usage_arg:"[-kill]"
      ~init:(Fn.const false)
      ~flags:[
        Command.Flag.noarg_acc "-kill" (Fn.const true)
          ~doc:" send SIGKILL instead of SIGTERM"
      ]
      ~final:(fun really_kill anons ->
        assert_no_anons anons;
        if really_kill then Signal.kill else Signal.term)
  in
  Command.group ~summary:(summary ~verb:"manage") [
    ("start",   base_cmd ~verb:"start" start_main);
    ("stop",    stop_cmd ~verb:"stop" stop_main);
    ("restart", stop_cmd ~verb:"restart" restart_main);
    ("status",  base_cmd ~verb:"check status of" status_main);
  ]

