(******************************************************************************
 *                             Core-extended                                  *
 *                                                                            *
 * Copyright (C) 2008- Jane Street Holding, LLC                               *
 *    Contact: opensource@janestreet.com                                      *
 *    WWW: http://www.janestreet.com/ocaml                                    *
 *                                                                            *
 *                                                                            *
 * This library is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU Lesser General Public                 *
 * License as published by the Free Software Foundation; either               *
 * version 2 of the License, or (at your option) any later version.           *
 *                                                                            *
 * This library is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this library; if not, write to the Free Software        *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 *                                                                            *
 ******************************************************************************)

open Core.Std ;;

exception Connect_timeout ;;

let with_connection ?(timeout=(Time.Span.of_float 30.)) ~host ~port ~f () =
  Exn.protectx
    (Unix.socket ~domain:Unix.PF_INET ~kind:Unix.SOCK_STREAM ~protocol:0)
    ~f:(fun sd ->
      Unix.set_nonblock sd;
      begin
        try
          Unix.connect sd ~addr:(Unix.get_sockaddr host port)
        with
        | Unix.Unix_error (Unix.EINPROGRESS, _, _) -> ()
        | e -> raise e
      end;
      let sfds =
        Unix.select
          ~read:[]
          ~write:[sd]
          ~except:[sd]
          ~timeout:(Time.Span.to_float timeout)
          ()
      in
      begin match sfds.Unix.Select_fds.except with
      | [] -> ()
      | sd :: [] ->

          Unix.unix_error
            (Unix.getsockopt_int sd Unix.SO_ERROR)
            "connect"
            (sprintf "%s:%d" host port)

      | _ -> assert false
      end;

      begin match sfds.Unix.Select_fds.write with
      | [] -> raise Connect_timeout
      | sd :: [] ->
          Unix.clear_nonblock sd;
          Exn.protectx
            (Unix.in_channel_of_descr sd,
             Unix.out_channel_of_descr sd)
          ~f:(fun (in_chan, out_chan) -> f in_chan out_chan)
          ~finally:(fun (in_chan, out_chan) ->
            try Out_channel.close out_chan with _e -> ();
            try In_channel.close in_chan with _e -> ()
          )

      | _ -> assert false
      end
    )
    ~finally:(fun sd -> try Unix.close sd with _e -> ())
;;

